from herre.models import Token
from typing import Protocol, runtime_checkable, Type
from pydantic import BaseModel


@runtime_checkable
class UserFetcher(Protocol):
    """A protocol for fetching users.

    A user fetcher is a class that is able to fetch a user from a token. It
    can be a parameter to the Herre class. The Herre class will then use the
    user fetcher to fetch the user from the token.


    """

    userModel: Type[BaseModel]

    async def afetch_user(self, token: Token) -> BaseModel:
        """Fetches the user from the token.


        Parameters:
        ___________
        token: Token
            The token to use to fetch the user.

        Returns:
        ________
        BaseModel
            The user as a pydantic model (will be userModel)

        """
        ...



from aiohttp import web
import asyncio
import webbrowser
import logging
from herre.grants.oauth2.errors import Oauth2RedirectError
from typing import Callable, Awaitable
from pydantic import BaseModel
from herre.models import TokenRequest
from typing import Optional, List
from pydantic import field_validator, Field

logger = logging.getLogger(__name__)

success_full_return = """
<!doctype html>

<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">

  <title>Sucessfull Login</title>
  <meta name="description" content="A success page for login">
  <meta name="author" content="SitePoint">

  <meta property="og:title" content="Sucessfull Login">
  <meta property="og:type" content="website">

</head>

<body>
    <h1>Successfull Login</h1>
    <p>You can close this window</p>
  <script type="text/javascript">
  window.close() ;
</script>
</body>
</html>
"""

failure_return = """
<!doctype html>

<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">

  <title>Sucessfull Login</title>
  <meta name="description" content="A success page for login">
  <meta name="author" content="SitePoint">

  <meta property="og:title" content="Sucessfull Login">
  <meta property="og:type" content="website">

</head>

<body>
    <h1>Unsucessfull login</h1>
    <p>You can close this window</p>
  <script type="text/javascript">
  window.close() ;
</script>
</body>
</html>
"""


async def find_free_port(
    host: str = "",
    selectable_ports: Optional[List[int]] = None,
    selectable_start=6000,
    selectable_end=7000,
) -> int:
    """Finds a free port

    This function will find a free port on the host machine.

    Parameters
    ----------
    host : str
        The host to check
    selectable_ports : Optional[List[int]]
        A list of ports to check

    Returns
    -------
    int
        The free port

    """

    if not selectable_ports:
        selectable_ports = list(range(selectable_start, selectable_end))

    for port in selectable_ports:
        try:
            server = await asyncio.start_server(lambda x, y: None, host=host, port=port)
            server.close()
            await server.wait_closed()
            return port
        except OSError:
            continue

    raise OSError("No free ports found")


def wrapped_qs_future(
    future: asyncio.Future, success_html: str, failure_html: str
) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Wraps a future in a webserver

    This is used to wrap a future in a webserver, so that the future can be resolved
    when the webserver is called. It is similar to an async partial function.

    Parameters
    ----------
    future : Future
        The future to wrap
    success_html : str
        The html to return when the future is resolved
    failure_html : str
        The html to return when the future is rejected

    """

    async def web_token_response(request: web.Request) -> web.Response:
        """The webserver response

        This function will resolve the future when called, and return a response
        to the caller.
        """
        try:
            future.set_result(request.path_qs)
            return web.Response(text=success_html, content_type="text/html")
        except Exception:
            return web.Response(text=failure_html, content_type="text/html")

    return web_token_response


def is_valid_port(port: int) -> bool:
    """Checks if a port is valid

    This function will check if a port is valid.

    Parameters
    ----------
    port : int
        The port to check

    Returns
    -------
    bool
        True if the port is valid, False otherwise

    """

    if not isinstance(port, int):
        return False

    if port < 0 or port > 65535:
        return False

    return True


class PortFinder(BaseModel):
    """A simple port finder"""

    fixed_port: Optional[int] = None
    selectable_ports: Optional[List[int]] = None
    selectable_start: int = 5000
    selectable_end: int = 6000

    @field_validator("selectable_ports")
    def check_ports(cls, v):
        if isinstance(v, list):
            for i in v:
                if not is_valid_port(i):
                    raise ValueError(f"Port {i} is not valid")

            return v

        if isinstance(v, int):
            if is_valid_port(v):
                return [v]
            else:
                raise ValueError(f"Port {v} is not valid")

        return []

    @field_validator("fixed_port", "selectable_start", "selectable_end")
    def check_fixed_port(cls, v):
        if v is not None:
            if is_valid_port(v):
                return v
            else:
                raise ValueError(f"Port {v} is not valid")

        return None

    async def afind_port(self, host: str = "") -> int:
        """Finds a free port

        This function will find a free port on the host machine.

        Parameters
        ----------
        host : str
            The host to check

        Returns
        -------
        int
            The free port

        """
        if self.fixed_port:
            return self.fixed_port

        return await find_free_port(
            host=host,
            selectable_ports=self.selectable_ports,
            selectable_start=self.selectable_start,
            selectable_end=self.selectable_end,
        )


class AioHttpServerRedirecter(BaseModel):
    """A simple webserver that will listen for a redirect from the OSF and return the path"""

    redirect_port: PortFinder = Field(
        default_factory=lambda: PortFinder(), description="The port to listen on"
    )
    redirect_timeout: int = 40
    redirect_host: str = "127.0.0.1"
    redirect_protocol: str = "http"
    redirect_path: str = "/"
    success_html: str = success_full_return
    failure_html: str = failure_return

    _chosen_port: Optional[int] = None

    @field_validator("redirect_port")
    def check_port(cls, v):
        if isinstance(v, str):
            v = int(v)

        if isinstance(v, int):
            return PortFinder(fixed_port=v)

        if isinstance(v, list):
            return PortFinder(selectable_ports=v)

        if isinstance(v, PortFinder):
            return v

        raise ValueError(f"Port {v} is not valid")

    async def aget_redirect_uri(self, token_request: TokenRequest) -> str:
        """Retrieves the redirect uri

        This function will retrieve the redirect uri from the RedirectWaiter.
        This function has to be implemented by the user.

        """

        self._chosen_port = await self.redirect_port.afind_port()

        return f"{self.redirect_protocol}://{self.redirect_host}:{self._chosen_port}{self.redirect_path}"

    async def astart(
        self,
        starturl: str,
    ) -> str:
        """Awaits a redirect

        This has to be implemented by a user

        """

        token_future = asyncio.get_event_loop().create_future()

        app = web.Application()

        app.router.add_get(
            "/",
            wrapped_qs_future(
                token_future,
                success_html=success_full_return,
                failure_html=failure_return,
            ),
        )

        port = self._chosen_port
        if not port:
            raise ValueError("Port not chosen")

        webserver_future = asyncio.wait_for(
            web._run_app(
                app,
                host=self.redirect_host,
                port=port,
                print=lambda x: logger.info(x),
                handle_signals=False,
            ),
            self.redirect_timeout,
        )

        webserver_task = asyncio.create_task(webserver_future)

        webbrowser.open(starturl)
        done, pending = await asyncio.wait(
            [token_future, webserver_task], return_when=asyncio.FIRST_COMPLETED
        )

        for tf in done:
            if tf == token_future:
                redirect_qs = tf.result()
            else:
                raise Oauth2RedirectError(
                    f"Webserver ended unexpectedly {str(tf.exception())}"
                )

        for task in pending:
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass

        if not redirect_qs:
            raise Oauth2RedirectError("Webserver")

        return redirect_qs



import aiohttp
from herre.grants.oauth2.base import BaseOauth2Grant
from herre.models import Token, TokenRequest

from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient

from herre.grants.oauth2.utils import build_token_url
from oauthlib.common import urldecode


class ClientCredentialsGrant(BaseOauth2Grant):
    """A grant that uses the client credentials flow"""

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches a token

        This function will fetch a token from the oauth2 provider,
        using the client credentials flow.


        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """
        scope = self.scope_delimiter.join(self.scopes)

        auth_client = BackendApplicationClient(
            client_id=self.client_id.get_secret_value(),
            scope=self.scope_delimiter.join(self.scopes),
        )

        token_url = build_token_url(self)

        body = auth_client.prepare_request_body(
            client_secret=self.client_secret.get_secret_value(),
            client_id=self.client_id.get_secret_value(),
        )

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }

        data = dict(urldecode(body))

        # Create an OAuth2 session for the OSF
        async with aiohttp.ClientSession(
            connector=(
                aiohttp.TCPConnector(ssl=self.ssl_context) if self.ssl_context else None
            ),
            headers=headers,
        ) as session:
            async with session.post(
                token_url,
                data=data,
                auth=aiohttp.BasicAuth(
                    self.client_id.get_secret_value(),
                    self.client_secret.get_secret_value(),
                ),
            ) as resp:
                text = await resp.text()

                auth_client.parse_request_body_response(text, scope=scope)

                token = auth_client.token
                return Token(**token)



import logging
from pydantic import BaseModel
from herre.models import TokenRequest
import urllib.parse

logger = logging.getLogger(__name__)


class MockRedirecter(BaseModel):
    """A simple webserver that will listen for a redirect from the OSF and return the path"""

    code: str
    """ The code to return """
    redirect_port: int = 6767
    redirect_timeout: int = 40
    redirect_host: str = "127.0.0.1"
    redirect_protocol: str = "http"
    redirect_path: str = "/"

    async def aget_redirect_uri(self, token_request: TokenRequest) -> str:
        """Retrieves the redirect uri

        This function will retrieve the redirect uri from the RedirectWaiter.
        This function has to be implemented by the user.

        """

        return f"{self.redirect_protocol}://{self.redirect_host}:{self.redirect_port}{self.redirect_path}"

    async def astart(
        self,
        auth_url: str,
    ) -> str:
        """Awaits a redirect

        This has to be implemented by a user

        """
        parsed_url = urllib.parse.urlparse(auth_url)
        query_params = urllib.parse.parse_qs(parsed_url.query)

        # Extract the redirect URI and state parameter
        redirect_uri_param = query_params.get("redirect_uri", [""])[0]
        state = query_params.get("state", [""])[0]

        if not redirect_uri_param:
            raise ValueError("Redirect URI not found in the authorization URL.")

        # Reconstruct the redirect URI with the code and state
        redirect_uri_components = list(urllib.parse.urlparse(redirect_uri_param))
        redirect_query_params = {"code": self.code}
        if state:
            redirect_query_params["state"] = state

        redirect_uri_components[4] = urllib.parse.urlencode(
            redirect_query_params
        )  # Set the query component
        return urllib.parse.urlunparse(redirect_uri_components)



from herre.fetcher.errors import UserFetchingError
from herre.grants.base import BaseGrant
from typing import Optional, runtime_checkable, Protocol
from pydantic import BaseModel, ConfigDict
from herre.models import Token, TokenRequest
from herre.grants.base import BaseGrantProtocol
import logging
import asyncio
from herre.fetcher.models import UserFetcher

logger = logging.getLogger(__name__)


class User(BaseModel):
    """A user model"""

    id: str
    username: str


class StoredUser(BaseModel):
    """A wrapping cl"""

    user: User
    token: Token


@runtime_checkable
class UserStore(Protocol):
    """A protocol for a user store

    This protocol is implemented by the user store.
    It can be used to type hint a user store. This
    is used by the AutoLoginGrant to store the user.

    """

    async def aget_default_user(self) -> Optional[StoredUser]:
        """Gets the default user

        Returns
        -------
        Optional[StoredUser]
            A stored user, with the token and the user
        """
        ...

    async def aput_default_user(self, user: Optional[StoredUser]) -> None:
        """Stores the default user


        Parameters
        ----------
        user : Optional[StoredUser]
            A stored user, with the token and the user, if None, the user should
            be deleted

        """
        ...


@runtime_checkable
class AutoLoginWidget(Protocol):
    """A protocol for a login widget

    This protocol is implemented by the login widget.
    It can be used to type hint a login widget. This
    is used by the AutoLoginGrant to show the login widget
    and ask the user if they want to save the user.
    """

    async def ashould_we_save(self, store: StoredUser) -> bool:
        """Should ask the user if we should save the user"""
        ...


class AutoLoginGrant(BaseGrant):
    """A grant that uses a Qt login screen to authenticate the user.

    The user is presented with a login screen that allows them to select a user
    from a list of previously logged in users. If the user is not in the list,
    they can click a button to start the login flow.


    """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    store: UserStore
    """this is the login widget (protocol)"""

    fetcher: UserFetcher

    widget: AutoLoginWidget
    """this is the login widget (protocol)"""

    grant: BaseGrantProtocol
    """The grant to use for the login flow."""

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches the token

        This function will only delegate to the grant if the user has not
        previously logged in (aka there is no token in the storage) Or if the
        force_refresh flag is set.

        Args:
            force_refresh (bool, optional): _description_. Defaults to False.

        Raises:
            e: _description_

        Returns:
            Token: _description_
        """

        try:
            if request.context.get("delete_active", True):
                await self.store.aput_default_user(None)

            if request.context.get("allow_auto_login", True):
                stored_user = await self.store.aget_default_user()
                if stored_user:
                    # Lets check if the token is still valid

                    try:
                        user = await self.fetcher.afetch_user(stored_user.token)
                        await self.store.aput_default_user(
                            StoredUser(user=user.dict(), token=stored_user.token)  # type: ignore # we dict here to ensure the serialization works
                        )
                        return stored_user.token
                    except UserFetchingError:
                        # The token is not valid anymore
                        token = await self.grant.afetch_token(request)
                        user = await self.fetcher.afetch_user(token)
                        await self.store.aput_default_user(
                            StoredUser(
                                user=user.dict(), token=token  # type: ignore# we dict here to ensure the serialization works
                            )
                        )
                        return token

                    # This time with a refresh

            # We are skipping the widget and just fetching the token
            token = await self.grant.afetch_token(request)
            user = await self.fetcher.afetch_user(token)

            new_store = StoredUser(user=user.dict(), token=token)  # type: ignore # we dict here to ensure the serialization works
            should_we_save = await self.widget.ashould_we_save(new_store)
            if should_we_save:
                await self.store.aput_default_user(new_store)
            else:
                await self.store.aput_default_user(None)

            return token

        except asyncio.CancelledError as e:
            raise e

        except Exception as e:
            logger.error(e, exc_info=True)
            raise e



from herre.grants.base import BaseGrant
from pydantic import ConfigDict, SecretStr, Field
from typing import List
import ssl
import certifi
from ssl import SSLContext


class BaseOauth2Grant(BaseGrant):
    """A base class for oauth2 grants."""
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")
    base_url: str
    """The base url to use for the grant"""

    client_id: SecretStr = SecretStr("")
    """The client id to use for the grant"""
    client_secret: SecretStr = SecretStr("")  #
    """The client secret to use for the grant"""
    scopes: List[str] = Field(default_factory=lambda: list(["openid"]))
    """The scopes to use for the grant"""
    authorize_path: str = "authorize"
    """The authorize path to use for the grant (relative to the base url)"""
    refresh_path: str = "token"
    """The refresh path to use for the grant (relative to the base url)"""
    token_path: str = "token"
    """The token path to use for the grant (relative to the base url)"""
    scope_delimiter: str = " "
    """The scope delimiter to use for the grant default is a space"""
    allow_insecure: bool = False
    """Whether to allow insecure connections"""
    append_trailing_slash: bool = True
    """Whether to append a trailing slash to the base url"""

    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where())
    )



""" Herre

Herre is a Python Library for easy integration with Token Based Authentication
in Python Applications, (for example using OAuth2). It is designed to be
extensible, and easy to use, and is built for the async world, while maintaining
sync compatibility.

Herre is build on top of aiohttp, and uses pydantic for data validation.


"""

from .herre import Herre, current_herre
from .builders import github_desktop

__all__ = ["Herre", "current_herre", "github_desktop"]



import aiohttp
from oauthlib.oauth2 import WebApplicationClient
import logging
from herre.grants.oauth2.utils import (
    build_authorize_url,
    build_token_url,
)
from .base import BaseOauth2Grant
from typing import Awaitable
from herre.models import Token, TokenRequest
from typing import Protocol, runtime_checkable
from oauthlib.common import generate_token, urldecode


logger = logging.getLogger(__name__)


@runtime_checkable
class Redirecter(Protocol):
    """A protocol for a from oauthlib.common import generate_tokenedirect waiter"""

    async def aget_redirect_uri(self, token_request: TokenRequest) -> str:
        """Retrieves the redirect uri

        This function will retrieve the redirect uri from the RedirectWaiter.
        This function has to be implemented by the user.

        """

    def astart(
        self,
        starturl: str,
    ) -> Awaitable[str]:
        """Awaits a redirect

        This has to be implemented by a user, and should
        return the path of the redirect (with the code)

        Parameters
        ----------
        starturl : str
            The url to start the redirect from

        Returns
        -------
        Awaitable[str]
            The path of the redirect (with the code)

        """
        ...


class AuthorizationCodeGrant(BaseOauth2Grant):
    """A grant that uses the authorization code flow

    This grant will create an AuthorizationCodeGrant, and use it to fetch a token.



    """

    redirecter: Redirecter
    """ A simple webserver that will listen for a redirect from the OSF and return the path """

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetch Token

        This function will fetch a token from the oauth2 provider,
        using the authorization code flow. It will retrieve the redirect_uri from the redirecter,
        and use that as the redirect_uri, it will then build an authorization url, and delegate the
        redirect to the RedirectWaiter. When the redirecter has received the redirect, it will
        return the code to this function, which will then use the code to fetch a token.


        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """

        state = generate_token()
        scope = self.scope_delimiter.join(self.scopes)
        redirect_uri = await self.redirecter.aget_redirect_uri(request)

        web_app_client = WebApplicationClient(
            self.client_id.get_secret_value(),
            scope=scope,
        )

        auth_url = web_app_client.prepare_request_uri(
            build_authorize_url(self),
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
        )

        web_app_client = WebApplicationClient(
            self.client_id.get_secret_value(),
            scope=self.scope_delimiter.join(self.scopes),
        )

        authorization_response = await self.redirecter.astart(
            auth_url,
        )

        web_app_client.parse_request_uri_response(
            str(authorization_response), state=state
        )
        code = web_app_client.code

        token_url = build_token_url(self)

        body = web_app_client.prepare_request_body(
            code=code,
            redirect_uri=redirect_uri,
            client_secret=self.client_secret.get_secret_value(),
        )

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }

        data = dict(urldecode(body))

        # Create an OAuth2 session for the OSF
        async with aiohttp.ClientSession(
            connector=(
                aiohttp.TCPConnector(ssl=self.ssl_context) if self.ssl_context else None
            ),
            headers=headers,
        ) as session:
            async with session.request(
                "POST",
                token_url,
                data=data,
            ) as resp:
                text = await resp.text()

                web_app_client.parse_request_body_response(text, scope=scope)

                token = web_app_client.token
                return Token(**token)



from herre.grants.base import BaseGrant
from herre.models import Token, TokenRequest


class StaticGrant(BaseGrant):
    """A grant that uses a static token

    THis grant will always return the same token.
    It is useful for testing.

    """

    token: Token

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches a token

        This function will return the token that was passed to the constructor.

        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """
        return self.token



from herre.grants.auto_login import (
    StoredUser,
)
from typing import Dict, Optional

import json
from pydantic import BaseModel
from fakts import Fakts
import logging
from qtpy import QtCore

logger = logging.getLogger(__name__)


class OrderDefaults(BaseModel):
    """A model for the default user storage

    It is used to store the default user for the fakts
    key.

    """

    default_user: Dict[str, StoredUser] = {}


class FaktsQtStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    settings: QtCore.QSettings
    default_user_key: str = "default_user_fakts"
    fakts: Fakts
    fakts_key: str

    async def aput_default_user(self, user: Optional[StoredUser]) -> None:
        """Puts the default user

        This method stores the user under a key that it retrieves from
        the fakts. This is done to ensure that the user is only stored
        for the correct fakts. (E.g. when the corresponding endpoint
        server changes)


        Parameters
        ----------
        user : StoredUser | None
            A stored user, with the token and the user, if None is provided
            the user is deleted

        """
        key = await self.fakts.aget(self.fakts_key)
        un_storage = self.settings.value(self.default_user_key, None)
        if not un_storage:
            storage = OrderDefaults()
        else:
            try:
                storage = OrderDefaults(**json.loads(un_storage))
            except Exception as e:
                logger.debug(e, exc_info=True)
                storage = OrderDefaults()

        if user is None:
            if key in storage.default_user:
                del storage.default_user[key]
        else:
            storage.default_user[key] = user

        self.settings.setValue(self.default_user_key, storage.json())

    async def aget_default_user(self) -> Optional[StoredUser]:
        """Gets the default user

        Gets the default user for the fakts key. If no user is stored
        for the fakts key, None is returned.

        Returns
        -------
        Optional[StoredUser]
            A stored user, with the token and the user
        """
        key = await self.fakts.aget(self.fakts_key)
        un_storage = self.settings.value(self.default_user_key, None)
        if not un_storage:
            return None
        try:
            storage = OrderDefaults(**json.loads(un_storage))
            if key in storage.default_user:
                return storage.default_user[key]
        except Exception as e:
            logger.warning(e, exc_info=True)
            return None

        return None

    class Config:
        """pydantic config"""

        arbitrary_types_allowed = True



from herre import Herre
from typing import Optional, List
from herre.grants.oauth2.authorization_code import AuthorizationCodeGrant
from herre.grants.oauth2.redirecters import AioHttpServerRedirecter


def github_desktop(
    client_id: str, client_secret: str, scopes: Optional[List[str]] = None
) -> Herre:
    """Creates a Herre instance that can be used to login locally to github

    This function will create a Herre instance that can be used to login locally to github.
    It will use the authorization code grant, and a aiohttp server redirecter.

    Parameters
    ----------
    client_id : str
        The client id to use
    client_secret : str
        The client secret to use
    scopes : Optional[List[str]], optional
        The scopes to use, by default None

    Returns
    -------
    Herre
        The Herre instance
    """
    if scopes is None:
        scopes = []

    return Herre(
        grant=AuthorizationCodeGrant(
            base_url="https://github.com/login/oauth",
            token_path="access_token",
            client_id=client_id,  # type: ignore
            client_secret=client_secret,  # type: ignore
            scopes=scopes,
            redirecter=AioHttpServerRedirecter(),
            append_trailing_slash=False,  # github does not like trailing slashes
        )
    )



""" Fakts Contrib Package

This package contains contrib grants and the grant registry
that allows for the fakts grant to build the correct grant
from the fakts.
    
"""

from .grant import FaktsGrant, GrantRegistry

__all__ = ["FaktsGrant", "GrantRegistry"]



from abc import abstractmethod

from pydantic import BaseModel, ConfigDict
from herre.models import Token, TokenRequest
from typing import Protocol, runtime_checkable
import logging

logger = logging.getLogger(__name__)


@runtime_checkable
class BaseGrantProtocol(Protocol):
    """The base grant protocol

    This protocol is implemented by all grants.
    It can be used to type hint a grant.

    """

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches a token

        This function will fetch a token from the grant.
        This function is async, and should be awaited

        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """
        ...


class BaseGrant(BaseModel):
    """The base grant class

    This class is the base class for all grants.
    It is a pydantic model, and can be used as such.
    It also implements the BaseGrantProtocol, which can be used to type hint
    a grant.

    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @abstractmethod
    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches a token

        This function will fetch a token from the grant.
        This function is async, and should be awaited

        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """
        raise NotImplementedError("Implement afetch_token")



""" Redirecters for OAuth2. 

This module contains redirecters that can be used by the herre
authorization code grant to redirect the user to the authorization
server, and to receive the code from the redirect uri.

We provide a few redirecters, but you can also create your own
redirecter, by implementing the RedirecterProtocol.





"""

from .mock import MockRedirecter
from .aiohttp_server import AioHttpServerRedirecter

__all__ = ["MockRedirecter", "AioHttpServerRedirecter"]



from qtpy import QtWidgets

from koil.qt import qt_to_async

from herre.grants.auto_login import (
    StoredUser,
)


class ShouldWeSaveThisAsDefault(QtWidgets.QDialog):
    """A dialog that asks the user if they want to save the user as the default user."""

    def __init__(self, stored: StoredUser, *args, **kwargs) -> None:
        """Creates a new ShouldWeSaveThisAsDefault dialog"""
        super().__init__(*args, **kwargs)
        self.setWindowTitle(f"Hi {stored.user.username}")

        self.qlabel = QtWidgets.QLabel(
            "Do you want to save yourself as the default user?"
        )

        vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(vlayout)

        vlayout.addWidget(self.qlabel)

        hlayout = QtWidgets.QHBoxLayout()
        vlayout.addLayout(hlayout)

        self.yes_button = QtWidgets.QPushButton("Yes")
        self.no_button = QtWidgets.QPushButton("No")

        self.yes_button.clicked.connect(self._on_yes)
        self.no_button.clicked.connect(self._on_no)

        self.stored = stored

        hlayout.addWidget(self.yes_button)
        hlayout.addWidget(self.no_button)

    def _on_yes(self) -> None:
        self.accept()

    def _on_no(self) -> None:
        self.reject()


class AutoLoginWidget(QtWidgets.QWidget):
    """A Qt widget for auto login.

    This widget can be used by the AutoLoginGrant to show the login widget
    and ask the user if they want to save the user.

    """

    def __init__(self, *args, **kwargs) -> None:
        """Creates a new AutoLoginWidget"""
        super().__init__(*args, **kwargs)

        self.ashould_we = qt_to_async(self._should_we, autoresolve=True)

    def _should_we(self, stored: StoredUser) -> bool:
        dialog = ShouldWeSaveThisAsDefault(stored, parent=self)
        dialog.exec_()
        return dialog.result() == QtWidgets.QDialog.Accepted

    async def ashould_we_save(self, store: StoredUser) -> bool:
        """Should ask the user if we should save the user"""
        return await self.ashould_we(store)



from herre.grants.base import BaseGrant, BaseGrantProtocol
from herre.models import Token, TokenRequest
import os
from typing import Optional
import pydantic
import datetime
import logging
import json

logger = logging.getLogger(__name__)


class CacheFile(pydantic.BaseModel):
    """Cache file model"""

    token: Token
    created: datetime.datetime
    hash: str = ""


class CacheGrant(BaseGrant):
    """Grant for caching data, caches the data of the its child grant in a file,
    if that file exists, and it is not expired, it will be used instead of delegating
    to the child grant."""

    grant: BaseGrantProtocol = pydantic.Field(..., description="The grant to cache")
    """The grant to cache"""
    cache_file: str = ".fakts_cache.json"
    """The cache file to use"""
    hash: str = pydantic.Field(
        default_factory=lambda: "",
        description="Validating against the hash of the config",
    )
    """The hash of the config to validate against"""
    expires_in: Optional[int]
    """The expiration time of the cache"""

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches a token

        This function will delegate to the child grant if the cache is expired or
        does not exist.

        Additionally, it will check the hash of the config, and the expiration data
        if it does not match, it will delegate to the child grant.

        Token Request Parameters:
        -------------------------
        allow_cache: bool
            Whether to allow the cache to be used

        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token
        """

        cache = None

        if os.path.exists(self.cache_file):
            with open(self.cache_file, "r") as f:
                x = json.load(f)
                try:
                    cache = CacheFile(**x)

                    if self.expires_in:
                        if (
                            cache.created + datetime.timedelta(seconds=self.expires_in)
                            < datetime.datetime.now()
                        ):
                            cache = None

                    if cache and self.hash and cache.hash != self.hash:
                        cache = None

                except pydantic.ValidationError as e:
                    logger.error(f"Could not load cache file: {e}. Ignoring it")

        if cache is None or not request.context.get("allow_cache", True):
            token = await self.grant.afetch_token(request)
            cache = CacheFile(
                token=token,
                created=datetime.datetime.now(),
            )

        with open(self.cache_file, "w") as f:
            json.dump(json.loads(cache.json()), f)

        return cache.token



from pydantic import BaseModel, Field
import ssl
import certifi
import aiohttp
from .models import Token
import logging
from .errors import UserFetchingError
from typing import Type

logger = logging.getLogger(__name__)


class UserinfoUserFetcher(BaseModel):
    """A user fetcher that fetches the user from an userinfo endpoint.

    This fetcher uses the userinfo endpoint to fetch the user. It uses the access token to
    authenticate itself to the userinfo endpoint.

    You can specify the model to use for the user. This model will be used to parse the answer
    from the userinfo endpoint. The model should be a pydantic model.


    """

    userModel: Type[BaseModel] = Field(
        description="The model to use for the user",
    )
    """ The model to use for the user"""
    userinfo_endpoint: str = Field(
        default="https://localhost:8000/userinfo",
        description="The endpoint to fetch the user from",
    )
    """ The endpoint to fetch the user from"""
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    async def afetch_user(self, token: Token) -> BaseModel:
        """Fetches the user from the userinfo endpoint.

        Parameters:
        ___________
        token: Token
            The token to use to authenticate to the userinfo endpoint.

        Returns:
        ________
        BaseModel
            The user as a pydantic model (will be userModel)

        """
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
            headers={"Authorization": f"Bearer {token.access_token}"},
        ) as session:
            async with session.get(
                f"{self.userinfo_endpoint}",
            ) as resp:
                if resp.status == 200:
                    try:
                        data = await resp.json()
                        return self.userModel(**data)
                    except Exception as e:
                        logger.error(f"Malformed answer: {data}")
                        raise UserFetchingError("Malformed Answer") from e

                else:
                    raise UserFetchingError(
                        "Error! Coud not retrieve on the endpoint. Maybe your token is invalid? Or you forgot to add the `openid` scope?"
                    )



""" Basic types for the herre library

This module contains the basic types for the herre library.

"""

from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field
import time


class TokenRequest(BaseModel):
    """A token request

    A token request is initiated by the client and contains all the information
    needed to request a token from the server.

    Grants can inspect the request and decide whether to handle it or not.
    Additionailly, they can modify the request before it is sent to the next grant.

    """

    is_refresh: bool = False
    """Whether this is a refresh request"""
    context: Dict[str, Any]
    """The context of the request"""


class Token(BaseModel):
    """A Token

    A token object contains all the information about a token.
    It mimics the oauthlib.oauth2.rfc6749.tokens.OAuthToken class.
    However, you can use it with any grant, not just oauth2 grants.
    As access_token is the only required field, you can use it as a simple
    bearer token.

    """

    access_token: str
    scope: Optional[List[str]] = Field(default_factory=list)
    refresh_token: Optional[str] = None
    expires_in: Optional[float] = None
    expires_at: Optional[float] = None
    token_type: Optional[str] = None

    def is_expired(self) -> bool:
        """Checks if the token is expired"""
        if self.expires_at:
            return self.expires_at < int(time.time())
        return False



from typing import Any, Dict, List, Optional
from pydantic import BaseModel, SecretStr
from herre.grants.base import BaseGrant
from herre.fakts.registry import GrantType, GrantRegistry
from herre.grants.oauth2.base import BaseOauth2Grant
from oauthlib.oauth2.rfc6749.errors import InvalidClientError
from fakts import Fakts
from herre.models import Token, TokenRequest


class HerreFakt(BaseModel):
    """A fakt for the herre grant"""

    base_url: str
    name: str
    client_id: SecretStr
    client_secret: SecretStr
    grant_type: GrantType
    grant_kwargs: Dict[str, Any] = {}
    scopes: List[str]
    redirect_uri: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    timeout: int = 500
    no_temp: bool = False


class FaktsGrant(BaseOauth2Grant):
    """A grant that uses fakts to configure itself

    Parameters
    ----------
    fakts : Fakts
        The fakts instance to use
    base_url : Optional[str], optional
        The base url to use for the grant, by default None
    grant_registry : GrantRegistry, optional
        The grant registry to use, by default get_default_grant_registry()
    fakts_group : str
        The fakts group to use for the grant


    """

    fakts: Fakts
    """Fakts instance to use"""
    grant_registry: GrantRegistry
    """The grant registry to use"""

    base_url: Optional[str] = None  # type: ignore
    """The base url to use for the grant (overwrites the one from the fakt)"""
    fakts_group: str
    """The fakts group to use for the grant"""
    allow_reconfiguration_on_invalid_client: bool = True
    """Whether to allow reconfiguration on invalid client errors"""

    _configured = False
    _activegrant: Optional[BaseGrant] = None
    _old_fakt: Dict[str, Any] = {}

    def configure(self, fakt: HerreFakt) -> None:
        """Configures the grant

        Sets the active grant to the grant specified in the fakt.

        Parameters
        ----------
        fakt : HerreFakt
            The fakt to configure the grant with

        Raises
        ------
        ValueError
            If the grant_type is not supported


        """
        grant_class = self.grant_registry.get_grant_for_type(fakt.grant_type)

        self._activegrant = grant_class(**fakt.dict())

    async def afetch_token(self, request: TokenRequest) -> Token:
        """Fetches the token

        This function will delegete to the active grant.
        If the underlying fakts has changed, it will reconfigure the grant.

        Parameters
        ----------
        request : TokenRequest
            The token request to use

        Returns
        -------
        Token
            The token

        Raises
        ------
        InvalidClientError
            If the client is invalid and allow_reconfiguration_on_invalid_client is False

        """

        if self.fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await self.fakts.aget(self.fakts_group)
            self.configure(HerreFakt(**self._old_fakt))

        assert self._activegrant is not None, "Grant not configured"

        try:
            return await self._activegrant.afetch_token(request)
        except InvalidClientError as e:
            if self.allow_reconfiguration_on_invalid_client:
                self._old_fakt = await self.fakts.aget(
                    self.fakts_group, force_refresh=True
                )
                self.configure(HerreFakt(**self._old_fakt))
                return await self._activegrant.afetch_token(request)
            else:
                raise e



from pydantic import BaseModel
from qtpy import QtCore
from typing import Optional
from herre.grants.stored_login import StoredUser


class QtSettingsUserStore(BaseModel):
    """A user store that uses Qt settings to store the use"""

    settings: QtCore.QSettings
    default_user_key: str = "default_user"

    async def aput_default_user(self, user: Optional[StoredUser]) -> None:
        """Puts the default user

        Parameters
        ----------
        user : StoredUser | None
            A stored user, with the token and the user, if None is provided
            the user is deleted
        """
        self.settings.setValue(self.default_user_key, user.json() if user else None)

    async def aget_default_user(self) -> Optional[StoredUser]:
        """Gets the default user

        Returns
        -------
        Optional[StoredUser]
            A stored user, with the token and the user
        """
        un_default_user = self.settings.value(self.default_user_key, None)
        if not un_default_user:
            return None
        return StoredUser.parse_raw(un_default_user)



from typing import Dict, Callable

from pydantic import BaseModel, ConfigDict, Field
from herre.grants.base import BaseGrant
import logging
from enum import Enum

logger = logging.getLogger(__name__)


class GrantType(str, Enum):
    """The grant type"""

    CLIENT_CREDENTIALS = "client-credentials"
    """The client credentials grant"""
    AUTHORIZATION_CODE = "authorization-code"
    """The authorization code grant"""


GrantBuilder = Callable[..., BaseGrant]
#


class GrantRegistry(BaseModel):
    """A registry for grants.

    This registry is used to register grants. It is used by the fakts
    grant to build the correct grant from the fakts.


    """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    registered_grants: Dict[GrantType, GrantBuilder] = Field(default_factory=dict)

    def register_grant(self, grant_type: GrantType, grant: GrantBuilder) -> None:
        """Registers a grant.

        Parameters:
        ___________
        type: GrantType
            The type of the grant to register
        grant: Type[BaseGrant]
            The grant to register

        """
        if not self.registered_grants:
            self.registered_grants = {}
        self.registered_grants[grant_type] = grant

    def get_grant_for_type(self, grant_type: GrantType) -> GrantBuilder:
        """Gets the grant for a type.

        Parameters:
        ___________
        type: GrantType
            The type of the grant to get

        Returns:
        ________
        Type[BaseGrant]
            The grant for the type

        """
        return self.registered_grants[grant_type]



import asyncio
from typing import Optional, TypeVar
from herre.errors import HerreError, NoHerreFound
from herre.grants.base import BaseGrant
import os
import logging
from herre.models import Token, TokenRequest
import contextvars
from koil.composition import KoiledModel
from koil.helpers import unkoil
from herre.fetcher.models import UserFetcher
from pydantic import BaseModel

current_herre: contextvars.ContextVar[Optional["Herre"]] = contextvars.ContextVar(
    "current_herre", default=None
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class Herre(KoiledModel):
    """Herre is a client for Token authentication.

    It provides a unified, composable interface for token based authentication based on grant.
    A grant is a class that is able to retrieve a token. Importantly grants do not have to
    directly call the token endpoint. They can also use a cache or other means to retrieve the
    token.

    Herre is a context manager. This allows it both to provide itself as a singleton and handle
    the asynchronous interface of the grant. As well as providing a lock to ensure that only one
    request is happening at a time.

    Example:
        ```python
        from herre import Herre,
        from herre.grants.oauth2.client_credentials import ClientCredentialsGrant

        herre = Herre(
            grant=ClientCredentialsGrant(
                client_id="my_client_id",
                client_secret="my_client
                base_url="https://my_token_url",
            )
        )

        with herre:
            token = herre.get_token()
        ```

        or aync

        ```python
        from herre import Herre,
        from herre.grants.oauth2.client_credentials import ClientCredentialsGrant

        herre = Herre(
            grant=ClientCredentialsGrant(
                client_id="my_client_id",
                client_secret="my_client
                base_url="https://my_token_url",
            )
        )

        async with herre:
            token = await herre.get_token()
        ```

    """

    grant: BaseGrant
    fetcher: Optional[UserFetcher] = None
    max_retries: int = 1
    allow_insecure: bool = False
    scope_delimiter: str = " "
    auto_login: bool = True

    login_on_enter: bool = False
    logout_on_exit: bool = False
    entered: bool = False

    no_temp: bool = False

    _lock: Optional[asyncio.Lock] = None
    _token: Optional[Token] = None

    @property
    def token(self) -> Token:
        "The current token"
        assert (
            self._lock is not None
        ), "Please enter the context first to access this variable"
        assert self._token is not None, "No token fetched"
        return self._token

    async def aget_token(self, **kwargs) -> str:
        """Get an access token

        Will return an access token if it is already available or
        try to login depending on auto_login. The checking and potential retrieving will happen
        in a lock ensuring that not multiple requests are happening at the same time.

        Args:
            auto_login (bool, optional): Should we allow an automatic login. Defaults to True.

        Returns:
            str:  The access token
        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"

        async with self._lock:
            if not self._token or not self._token.access_token:
                await self.arequest_from_grant(TokenRequest(context=kwargs))

        assert self._token is not None, "We should have a token by now"
        return self._token.access_token

    async def arefresh_token(self, **kwargs) -> str:
        """Refresh the token

        Will cause the linked grant to refresh the token. Depending
        on the link logic, this might cause another login.

        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"

        async with self._lock:
            await self.arequest_from_grant(
                TokenRequest(is_refresh=True, context=kwargs)
            )
            assert self._token is not None, "We should have a token by now"
            return self._token.access_token

    async def arequest_from_grant(self, request: TokenRequest) -> Token:
        """Request a token from the grant

        You should not need to call this method directly. It is used internally
        to request a token from the grant, and will not directly acquire a lock
        (so multiple requests can happen at the same time, which is often not what
        you want).

        Parameters
        ----------
        request : TokenRequest
            The token request (contains context and whether it is a refresh request)

        Returns
        -------
        Token
            The token (with access_token, refresh_token, etc.)
        """
        potential_token = await self.grant.afetch_token(request)
        self._token = potential_token
        return self._token

    def get_token(self, **kwargs) -> str:
        """Get an access token

        Will return an access token if it is already available or
        try to login depending on auto_login. The checking and potential retrieving will happen
        in a lock ensuring that not multiple requests are happening at the same time.
        """
        return unkoil(self.aget_token, **kwargs)

    async def aget_user(self, **kwargs) -> BaseModel:  # TODO: Should be generic
        """Get the current user

        Will return the current user if a fetcher is available
        """
        assert (
            self._lock is not None
        ), "We were not initialized. Please enter the context first"
        assert self.fetcher is not None, "We have no fetcher available"
        if not self._token:
            raise HerreError("No token available")
        async with self._lock:
            if not self._token or not self._token.access_token:
                await self.arequest_from_grant(TokenRequest(context=kwargs))

        assert self._token is not None, "We should have a token by now"
        return await self.fetcher.afetch_user(self._token)

    async def __aenter__(self) -> "Herre":
        """Enters the context and logs in if needed"""
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if self.allow_insecure else "1"
        self._lock = asyncio.Lock()
        current_herre.set(self)
        if self.login_on_enter:
            await self.alogin()
        self.entered = True
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Exits the context and logs out if needed"""
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0"
        if self.logout_on_exit:
            await self.alogout()
        current_herre.set(None)

    def _repr_html_inline_(self) -> str:
        """Jupyter inline representation"""
        return f"<table><tr><td>auto_login</td><td>{self.auto_login}</td></tr></table>"



def get_current_herre() -> Herre:
    """Get the current herre instance"""
    herre = current_herre.get()

    if herre is None:
        raise NoHerreFound("No herre instance available")

    return herre



""" OAuth2 Grants

This module contains grants that can be used by the herre
class to acquire Tokens using the OAuth2 protocol.
"""

from .client_credentials import ClientCredentialsGrant
from .refresh import RefreshGrant
from .authorization_code import AuthorizationCodeGrant

__all__ = ["ClientCredentialsGrant", "RefreshGrant", "AuthorizationCodeGrant"]



""" Grants module

This module contains grants that can be used by the herre
class to acquire Tokens.

"""

from .meta import CacheGrant
from .base import BaseGrant
from .static import StaticGrant

__all__ = ["CacheGrant", "BaseGrant", "StaticGrant"]



"""
This module contains the meta grants, which are grants that are not
meant to be used directly, but rather serve as a base for other grants.



"""

from .cache import CacheGrant

__all__ = ["CacheGrant"]



from herre.models import TokenRequest
from koil.qt import qt_to_async, QtFuture
from qtpy.QtWebEngineWidgets import QWebEngineView
from qtpy import QtCore


class WebViewRedirecter(QWebEngineView):
    """A qt webengine view that can be used to login to a oauth2 provider

    This class is used to login to a oauth2 provider, when using the authorization code grant,
    but without a webserver. This class will open a webview, and wait for the user to login.
    When the user is logged in, the user will be redirected to the redirect_uri.
    This redirect_uri will be intercepted, and the code will be returned to the caller.


    """

    def __init__(
        self, *args, redirect_uri: str = "http://localhost:4893/", **kwargs
    ) -> None:
        """Creates a new WebViewRedirecter

        This function will create a new LoginWrapper, and initialize the webview.

        Parameters
        ----------
        redirect_uri : str
            The redirect_uri to intercept (you shoud own this uri)

        **kwargs
            The arguments to pass to the webview

        """
        super().__init__(*args, **kwargs)
        self.name = self.page()
        self.redirect_uri = redirect_uri
        self.future = None
        self.show_coro = qt_to_async(self.initialize)

    def initialize(self, future: QtFuture, auth_url: str) -> None:
        """Initializes the LoginWrapper

        This function will initialize the LoginWrapper, and open the webview.
        It will also set the future to resolve when the user is logged in.

        Parameters
        ----------
        future : QtFuture
            The future to resolve when the user is logged in
        auth_url : str
            The url to login to
        redirect_uri : str
            The redirect_uri to intercept
        """
        self.future = future

        self.load(QtCore.QUrl(auth_url))
        self.show()
        self.urlChanged.connect(self.on_urlChanged)

    def on_urlChanged(self, url: QtCore.QUrl) -> None:
        """Intercepts a url

        This function will intercept a url, and resolve the future when the url
        starts with the redirect_uri
        """
        url_string = bytes(url.toEncoded()).decode()
        if self.redirect_uri:
            if url_string.startswith(self.redirect_uri):
                if self.future:
                    self.future.resolve(url_string)
                    self.close()
                else:
                    pass

    async def astart(self, auth_url: str) -> str:
        """Starts the LoginWrapper

        This function will start the LoginWrapper, and return the code when the user is logged in.

        Parameters
        ----------
        auth_url : str
            The url to login to
        redirect_uri : str
            The redirect_uri to intercept

        Returns
        -------
        str
            The redirect_uri with the code
        """
        return await self.show_coro(auth_url)

    async def aget_redirect_uri(self, token_request: TokenRequest) -> str:
        """Retrieves the redirect uri

        This function will retrieve the redirect uri from the RedirectWaiter.
        This function has to be implemented by the user.

        """

        return self.redirect_uri



from pydantic import BaseModel, ConfigDict, Field
import ssl
import certifi
import aiohttp
from herre.fetcher.models import Token
import logging
from herre.fetcher.errors import UserFetchingError
from fakts import Fakts
from typing import Type

logger = logging.getLogger(__name__)


class FaktsUserFetcher(BaseModel):
    """The endpoint to fetch the user from"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    userModel: Type[BaseModel] = Field(
        description="The model to use for the user",
    )
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    fakts: Fakts
    fakts_key: str

    """ An ssl context to use for the connection to the endpoint"""

    async def afetch_user(self, token: Token) -> BaseModel:
        """Fetches the user from the endpoint

        Parameters
        ----------
        token : Token
            The token to use for the request

        Returns
        -------
        BaseModel
            The userModel filled with the data from the endpoint
        """
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
            headers={"Authorization": f"Bearer {token.access_token}"},
        ) as session:
            user_info_url = await self.fakts.aget(self.fakts_key)
            async with session.get(
                user_info_url + "/",
            ) as resp:
                if resp.status == 200:
                    try:
                        data = await resp.json()
                        return self.userModel(**data)
                    except Exception as e:
                        logger.error(f"Malformed answer: {data}", exc_info=True)
                        raise UserFetchingError("Malformed Answer") from e

                else:
                    try:
                        resp.raise_for_status()
                    except Exception as e:
                        logger.error(f"Could not fetch user: {e}", exc_info=True)
                        raise UserFetchingError(
                            f"Could not fetch user from {user_info_url}. Maybe your token is invalid? Or you forgot to add the `openid` scope?"
                        ) from e

        raise UserFetchingError("Could not fetch user")

